import tester.*;

// PROGRAM BY DESIGN: HOW TO DESIGN CLASSES (REACH_JAVA)
// copyright 2011 VIERA KRNANOVA PROULX
// ISSEP 2011 WORKSHOP

// Part 4: A union of classes with methods
// File:   Book4.java

/*
                         +---------+                         
                         | LibItem |                         
                         +---------+                         
                         +---------+                         
                             / \                             
                              |                              
          - - - - - - - - - - - - - - - - - - - - -           
          |                   |                   |          
  +---------------+    +--------------+    +--------------+  
  | Book5         |    | CD           |    | Magazine     |  
  +---------------+    +--------------+    +--------------+  
  | int catNo     |    | int catNo    |    | int catNo    |  
  | int dayDue    |    | int dayDue   |    | int dayDue   |  
  | String title  |    | String title |    | String title |  
  | String author |    | int noCDs    |    | int vol      |  
  +---------------+    +--------------+    | int no       |  
                                           +--------------+  
                                                                                                                        
*/

// to represent library items
interface LibItem {

  // is this library item overdue given today's date
  boolean isOverdue(int today);

  // compute the fine for this library item given today's date
  int computeFine(int today);
}

// to represent a book in a library
public class Book4 implements LibItem {
  int catNo;
  int dayDue;
  String title;
  String author;

  Book4(int catNo, int dayDue, String title, String author) {
    this.catNo = catNo;
    this.dayDue = dayDue;
    this.title = title;
    this.author = author;
  }

/*
  // ** DRAFT TEMPLATE ** Edit as needed.
  ??? mmm() {
    ... this.catNo ...   -- int
    ... this.dayDue ...  -- int
    ... this.title ...   -- String
    ... this.author ...  -- String
  }
*/

  // is this book overdue given today's date
  public boolean isOverdue(int today){
    return this.dayDue < today;
  }

  // compute the fine for this book given today's date
  // 10 cents per day up to $10 maximum
  public int computeFine(int today){
    if (this.isOverdue(today))
      return Math.min(1000, (today - this.dayDue) * 10);
    else
      return 0;
  }
}

// to represent a CD in a library
class CD implements LibItem {
  int catNo;
  int dayDue;
  String title;
  int noCDs;

  CD(int catNo, int dayDue, String title, int noCDs) {
    this.catNo = catNo;
    this.dayDue = dayDue;
    this.title = title;
    this.noCDs = noCDs;
  }

/*
  // ** DRAFT TEMPLATE ** Edit as needed.
  ??? mmm() {
    ... this.catNo ...   -- int
    ... this.dayDue ...  -- int
    ... this.title ...   -- String
    ... this.noCDs ...   -- int
  }
*/

  // is this cd overdue given today's date
  public boolean isOverdue(int today){
    return this.dayDue < today;
  }

  // compute the fine for this cd given today's date
  // 50 cents per day up to $5 per cd in the set maximum
  public int computeFine(int today){
    if (this.isOverdue(today))
      return Math.min(500 * this.noCDs, (today - this.dayDue) * 50);
    else
      return 0;
  }
}

// to represent a magazine in a library
class Magazine implements LibItem {
  int catNo;
  int dayDue;
  String title;
  int vol;
  int no;

  Magazine(int catNo, int dayDue, String title, int vol, int no) {
    this.catNo = catNo;
    this.dayDue = dayDue;
    this.title = title;
    this.vol = vol;
    this.no = no;
  }

/*
  // ** DRAFT TEMPLATE ** Edit as needed.
  ??? mmm() {
    ... this.catNo ...   -- int
    ... this.dayDue ...  -- int
    ... this.title ...   -- String
    ... this.vol ...     -- int
    ... this.no ...      -- int
  }
*/

  // is this magazine overdue given today's date
  public boolean isOverdue(int today){
    return this.dayDue < today;
  }

  // compute the fine for this cd given today's date
  // 30 cents per day up to $10
  public int computeFine(int today){
    if (this.isOverdue(today))
      return Math.min(1000, (today - this.dayDue) * 30);
    else
      return 0;
  }

}

//examples of data and tests
class ExamplesBook4 {
  ExamplesBook4 () {}

  Book4 b1 = new Book4(2233, 223, "RC", "DD");
  Book4 b2 = new Book4(2255, 228, "LPP", "StEx");

  CD cd1 = new CD(4433, 223, "French", 5);
  CD cd2 = new CD(4422, 229, "Aida", 4);

  Magazine ddj = new Magazine(7755, 228, "DDJ", 31, 2);
  Magazine cacm = new Magazine(7722, 224, "CACM", 38, 5);

  // test the method isOverdue
  boolean testIsOverdue(Tester t){
	  return
	  t.checkExpect(this.b1.isOverdue(225), true) &&
	  t.checkExpect(this.b2.isOverdue(225), false) &&
	  
	  t.checkExpect(this.cd1.isOverdue(225), true) &&
	  t.checkExpect(this.cd2.isOverdue(225), false) &&
	  
	  t.checkExpect(this.ddj.isOverdue(225), false) &&
	  t.checkExpect(this.cacm.isOverdue(225), true);	  
  }

  // test the method computeFine
  // 10 cents per day up to $10 maximum
  boolean testComputeFine1(Tester t){
	  return

	  // 10 cents per day up to $10 maximum
	  t.checkExpect(this.b1.computeFine(225), 20) &&
	  t.checkExpect(this.b2.computeFine(225), 0) &&
	  t.checkExpect(this.b2.computeFine(400), 1000) &&

	  //50 cents per day up to $5 per cd in the set maximum
	  t.checkExpect(this.cd1.computeFine(225), 100) &&
	  t.checkExpect(this.cd2.computeFine(225), 0) &&
	  t.checkExpect(this.cd2.computeFine(400), 2000) &&

	  //30 cents per day up to $10
	  t.checkExpect(this.ddj.computeFine(225), 0) &&
	  t.checkExpect(this.cacm.computeFine(225), 30) &&
	  t.checkExpect(this.cacm.computeFine(400), 1000);
  }
  
  // run the tests
  public static void main(String[] argv){
	  ExamplesBook4 eb4 = new ExamplesBook4();
	  Tester.runFullReport(eb4);
  }
}
