import tester.Tester;

// PROGRAM BY DESIGN: HOW TO DESIGN CLASSES (REACH_JAVA)
// copyright VIERA K. PROULX



// Part 5a A self-referential union of classes: representing lists
//         with methods
// File:   Book6.java

/*
            +----------+                
            | ILoBook2 |<-------------+
            +----------+              |
            +----------+              |
                / \                   |
                ---                   |
                 |                    |
       -------------------            |
       |                 |            |
 +-----------+    +---------------+   |
 | MTLoBook2 |    | ConsLoBook2   |   |
 +-----------+    +---------------+   |
 +-----------+  +-| Book6 first   |   |
                | | ILoBook2 rest |---+
                | +---------------+ 
                |
                v
       +---------------+
       | Book6         |
       +---------------+
       | String title  |
       | String author |
       | int year      |
       +---------------+
*/

// to represent a list of books
interface ILoBook2 {

  // count the number of items in this list
  int count();

  // is there a book in this list published before the given year
  boolean hasPublishedBefore(int year);

  // produce a list of all books in this list 
  // published before the given year 
  ILoBook2 publishedBefore(int year);

  // produce a list sorted by the titles from this list of books
  ILoBook2 sortByTitle();

  // insert the given book into this sorted list of books
  ILoBook2 insertByTitle(Book6 b);
}

// to represent an empty list of books
class MTLoBook2 implements ILoBook2 {

  MTLoBook2() {
  }

/*
  // ** TEMPLATE ** Edit as needed.
  ...
*/

  // count the number of items in this list
  public int count(){ return 0; }

  // is there a book in this list published before the given year
  public boolean hasPublishedBefore(int year){ return false; }

  // produce a list of all books in this list 
  // published before the given year 
  public ILoBook2 publishedBefore(int year){ return this; }

  // produce a list sorted by the titles from this list of books
  public ILoBook2 sortByTitle(){ return this; }

  // insert the given book into this sorted list of books
  public ILoBook2 insertByTitle(Book6 b){ 
    return new ConsLoBook2(b, this);
  }
}

// to represent a nonempty list of Books
class ConsLoBook2 implements ILoBook2 {
  Book6 first;
  ILoBook2 rest;

  ConsLoBook2(Book6 first, ILoBook2 rest) {
    this.first = first;
    this.rest = rest;
  }
  
/* TEMPLATE: (built as new methods are added)
  FIELDS:
  ... this.first ...                          -- Book
  ... this.rest ...                           -- ILoBook
  
  METHODS:
  ... this.count() ...                   -- int
  ... this.hasPublishedBefore(year) ...  -- boolean
  ... this.publishedBefore(year) ...     -- ILoBook
  ... this.sortByTitle() ...             -- ILoBook (sorted)
  ... this.insertByTitle(Book) ...       -- ILoBook (sorted)
  
  METHODS FOR FIELDS:
  ... this.rest.count() ...                   -- int
  ... this.rest.hasPublishedBefore(year) ...  -- boolean
  ... this.rest.publishedBefore(year) ...     -- ILoBook
  ... this.rest.sortByTitle() ...             -- ILoBook (sorted)
  ... this.rest.insertByTitle(Book) ...       -- ILoBook (sorted)
*/
  
  // count the number of items in this list
  public int count(){ 
  /* TEMPLATE:
    ... this.first ...           -- Book
    ... this.rest ...            -- ILoBook
    ... this.rest.count() ...    -- int
  */
    return 1 + this.rest.count(); 
  }

  // is there a book in this list published before the given year
  public boolean hasPublishedBefore(int year){ 
  /* TEMPLATE:
    ... this.first ...                            -- Book
    ... this.rest ...                             -- ILoBook
    ... this.rest.count() ...                     -- int
    ... this.rest.hasPublishedBefore(year) ...    -- boolean
  */
    return this.first.publishedBefore(year) ||
           this.rest.hasPublishedBefore(year); 
  }

  // produce a list of all books in this list
  // published before the given year 
  public ILoBook2 publishedBefore(int year){ 
  /* TEMPLATE:
    ... this.first ...                            -- Book
    ... this.rest ...                             -- ILoBook
    ... this.rest.count() ...                     -- int
    ... this.rest.hasPublishedBefore(year) ...    -- boolean
    ... this.rest.publishedBefore(year) ...       -- ILoBook
  */
    if (this.first.publishedBefore(year))
      return new ConsLoBook2(this.first, 
                            this.rest.publishedBefore(year));
    else
      return this.rest.publishedBefore(year);
  }

  // produce a list sorted by the titles from this list of books
  public ILoBook2 sortByTitle(){  
  /* TEMPLATE:
    ... this.first ...                           -- Book
    ... this.rest ...                            -- ILoBook
    ... this.rest.count() ...                    -- int
    ... this.rest.hasPublishedBefore(year) ...   -- boolean
    ... this.rest.publishedBefore(year) ...      -- ILoBook
    ... this.rest.sortByTitle() ...              -- ILoBook (sorted)
  */
    return this.rest.sortByTitle().insertByTitle(this.first);
  }

  // insert the given book into this sorted list of books
  public ILoBook2 insertByTitle(Book6 b){
  /* TEMPLATE:
    ... this.first ...                          -- Book
    ... this.rest ...                           -- ILoBook
    ... this.rest.count() ...                   -- int
    ... this.rest.hasPublishedBefore(year) ...  -- boolean
    ... this.rest.publishedBefore(year) ...     -- ILoBook
    ... this.rest.sortByTitle() ...             -- ILoBook (sorted)
    ... this.rest.insertByTitle(Book) ...       -- ILoBook (sorted)
  */
    if (this.first.beforeByTitle(b))
      return new ConsLoBook2(this.first, this.rest.insertByTitle(b));
    else
      return new ConsLoBook2(b, this);
  }

}

// to represent a book
public class Book6 {
  String title;
  String author;
  int year;

  Book6(String title, String author, int year) {
    this.title = title;
    this.author = author;
    this.year = year;
  }
/* TEMPLATE
  FIELDS:
    ... this.title ...    -- String
    ... this.author ...   -- String
    ... this.year ...     -- int
  
  METHODS:
  ... publishedBefore(int) ...   -- boolean
  ... beforeByTitle(Book6) ...   -- boolean
*/

  // was this book published before the given year
  boolean publishedBefore(int year){ 
    return this.year < year;
  }

  // is the title of this book before the title of the given book
  boolean beforeByTitle(Book6 that){
    return this.title.compareTo(that.title) < 0;
  }
}

// examples of data and tests
class ExamplesBook6{
  ExamplesBook6() {}

  Book6 b1 = new Book6("RC", "DD", 1796);
  Book6 b2 = new Book6("LPP", "StEx", 1942);
  Book6 b3 = new Book6("HtDP", "FFFK", 2001);

  ILoBook2 mtbooks = new MTLoBook2();
  ILoBook2 books1 = new ConsLoBook2(this.b1, 
                   new ConsLoBook2(this.b2, 
                   this.mtbooks));
  ILoBook2 books2 = new ConsLoBook2(this.b3, this.books1);

  ILoBook2 bSorted = new ConsLoBook2(this.b3, 
                    new ConsLoBook2(this.b1, 
                    this.mtbooks));

  ILoBook2 allSorted = new ConsLoBook2(this.b3, 
                      new ConsLoBook2(this.b2, 
                      new ConsLoBook2(this.b1, 
                      this.mtbooks)));



  // tests for the method count for ILoBook classes
  boolean testCount(Tester t){
	  return
	  t.checkExpect(this.mtbooks.count(), 0) &&
	  t.checkExpect(this.books2.count(), 3);
  }

  // tests for the method hasPublishedBefore  for ILoBook classes
  boolean testHasPublishedBefore(Tester t){
	  return
	  t.checkExpect(this.mtbooks.hasPublishedBefore(2006), false) &&
	  t.checkExpect(this.books1.hasPublishedBefore(1700), false) &&
	  t.checkExpect(this.books2.hasPublishedBefore(1950), true);
  }

  // tests for the method publishedBefore  for ILoBook classes
  boolean testPublishedBefore(Tester t){
	  return
	  t.checkExpect(this.mtbooks.publishedBefore(2006), this.mtbooks) &&
	  t.checkExpect(this.books1.publishedBefore(1700), this.mtbooks) &&
	  t.checkExpect(this.books2.publishedBefore(1950), this.books1);
  }

  // tests for the method beforeByTitle in the class Book
  boolean testBeforeByTitle(Tester t){
	  return
	  t.checkExpect(this.b2.beforeByTitle(this.b1), true) &&
	  t.checkExpect(this.b2.beforeByTitle(this.b3), false);
  }

  // tests for the method insertByTitle for ILoBook classes
  boolean testInsertByTitle(Tester t){
	  return
	  t.checkExpect(this.mtbooks.insertByTitle(this.b1),
			  new ConsLoBook2(this.b1, this.mtbooks)) &&
	  t.checkExpect(this.bSorted.insertByTitle(this.b2),
              new ConsLoBook2(this.b3,
                new ConsLoBook2(this.b2,
				  new ConsLoBook2(this.b1,
				    this.mtbooks))));
  }


  // tests for the method sortByTitle for ILoBook classes
  boolean testSortByTitle(Tester t){
	  return
	  t.checkExpect(this.mtbooks.sortByTitle(), this.mtbooks) &&
	  t.checkExpect(this.books2.sortByTitle(), this.allSorted);
  }
  
  // run the tests
  public static void main(String[] argv){
	  ExamplesBook6 eb6 = new ExamplesBook6();
	  Tester.runFullReport(eb6);
  }  
}
