import tester.*;

// PROGRAM BY DESIGN: HOW TO DESIGN CLASSES (REACH_JAVA)
// copyright 2011 VIERA KRNANOVA PROULX
// ISSEP 2011 WORKSHOP


// Part 7: Abstracting over unions: abstract class
// File:   Book7.java

/*
                        +------------+                         
                        | LibItem2   |                         
                        +------------+ 
                        | int catNo  |
                        | int dayDue | 
                        +------------+                         
                             / \                             
                             ---                             
                              |                              
          -----------------------------------------           
          |                   |                   |          
  +---------------+    +--------------+    +--------------+  
  | Book7         |    | CD2          |    | Magazine2    |  
  +---------------+    +--------------+    +--------------+  
  | String title  |    | String title |    | String title |  
  | String author |    | int noCDs    |    | int vol      |  
  +---------------+    +--------------+    | int no       |  
                                           +--------------+  
                                                                                                                        
*/

// to represent library items
abstract class LibItem2 {
  int catNo;
  int dayDue;

  LibItem2(int catNo, int dayDue){
    this.catNo = catNo;
    this.dayDue = dayDue;
  }

  // is this book overdue given today's date
  boolean isOverdue(int today){
    return this.dayDue < today;
  }

  // compute the fine for this library item given today's date
  abstract int computeFine(int today);
}

// to represent a book in a library
class Book7 extends LibItem2 {
  String title;
  String author;

  Book7(int catNo, int dayDue, String title, String author) {
    super(catNo, dayDue);
    this.title = title;
    this.author = author;
  }

/* TEMPLATE 
  FIELDS:
    ... this.catNo ...
    ... this.dayDue ...
    ... this.title ...
    ... this.author ...
  
   METHODS:
   ... isOverdue(int) ...           -- boolean
   ... computeFine(int) ...         -- int
*/

  // compute the fine for this book given today's date
  // 10 cents per day up to $10 maximum
  int computeFine(int today){
    if (this.isOverdue(today))
      return Math.min(1000, (today - this.dayDue) * 10);
    else
      return 0;
  }
}

// to represent a CD in a library
class CD2 extends LibItem2 {
  String title;
  int noCDs;

  CD2(int catNo, int dayDue, String title, int noCDs) {
    super(catNo, dayDue);
    this.title = title;
    this.noCDs = noCDs;
  }

/* TEMPLATE 
  FIELDS:
    ... this.catNo ...
    ... this.dayDue ...
    ... this.title ...
    ... this.noCDs ...
  
   METHODS:
   ... isOverdue(int) ...           -- boolean
   ... computeFine(int) ...         -- int
*/

  // compute the fine for this cd given today's date
  // 50 cents per day up to $5 per cd in the set maximum
  int computeFine(int today){
    if (this.isOverdue(today))
      return Math.min(500 * this.noCDs, (today - this.dayDue) * 50);
    else
      return 0;
  }
}

// to represent a magazine in a library
class Magazine2 extends LibItem2 {
  String title;
  int vol;
  int no;

  Magazine2(int catNo, int dayDue, String title, int vol, int no) {
    super(catNo, dayDue);
    this.title = title;
    this.vol = vol;
    this.no = no;
  }

/* TEMPLATE 
  FIELDS:
    ... this.catNo ...
    ... this.dayDue ...
    ... this.title ...
    ... this.vol ...
    ... this.no ...
  
   METHODS:
   ... isOverdue(int) ...           -- boolean
   ... computeFine(int) ...         -- int
*/

  // compute the fine for this cd given today's date
  // 30 cents per day up to $10
  int computeFine(int today){
    if (this.isOverdue(today))
      return Math.min(1000, (today - this.dayDue) * 30);
    else
      return 0;
  }

}

//examples of data and tests
class ExamplesBook7 {
  ExamplesBook7 () {}

  Book7 b1 = new Book7(2233, 223, "RC", "DD");
  Book7 b2 = new Book7(2255, 228, "LPP", "StEx");

  CD2 cd1 = new CD2(4433, 223, "French", 5);
  CD2 cd2 = new CD2(4422, 229, "Aida", 4);

  Magazine2 ddj = new Magazine2(7755, 228, "DDJ", 31, 2);
  Magazine2 cacm = new Magazine2(7722, 224, "CACM", 38, 5);

  // test the method isOverdue
  boolean testIsOverdue(Tester t){
	  return
	  t.checkExpect(this.b1.isOverdue(225), true) &&
	  t.checkExpect(this.b2.isOverdue(225), false) &&

	  t.checkExpect(this.cd1.isOverdue(225), true) &&
	  t.checkExpect(this.cd2.isOverdue(225), false) &&

	  t.checkExpect(this.ddj.isOverdue(225), false) &&
	  t.checkExpect(this.cacm.isOverdue(225), true);
  }
 
  // test the method computeFine
  // 10 cents per day up to $10 maximum
  boolean testComputeFine(Tester t){
	  return
	  t.checkExpect(this.b1.computeFine(225), 20) &&
	  t.checkExpect(this.b2.computeFine(225), 0) &&
	  t.checkExpect(this.b2.computeFine(400), 1000) &&

  // 50 cents per day up to $5 per cd in the set maximum
	  t.checkExpect(this.cd1.computeFine(225), 100) &&
	  t.checkExpect(this.cd2.computeFine(225), 0) &&
	  t.checkExpect(this.cd2.computeFine(400), 2000) &&

  // 30 cents per day up to $10
	  t.checkExpect(this.ddj.computeFine(225), 0) &&
	  t.checkExpect(this.cacm.computeFine(225), 30) &&
	  t.checkExpect(this.cacm.computeFine(400), 1000);
  }
  
  // run the tests
  public static void main(String[] argv){
	  ExamplesBook7 eb7 = new ExamplesBook7();
	  Tester.runFullReport(eb7);
  }  
}
