import tester.*;

// PROGRAM BY DESIGN: HOW TO DESIGN CLASSES (REACH_JAVA)
// copyright 2011 VIERA KRNANOVA PROULX
// ISSEP 2011 WORKSHOP


// Part 8: A self-referential union of classes: representing lists
//         Motivating abstraction over functional behavior
// File:   Book8.java

/*
            +---------+                
            | ILoBook3 |<-------------+
            +---------+               |
            +---------+               |
                / \                   |
                ---                   |
                 |                    |
       -------------------            |
       |                 |            |
 +-----------+    +---------------+   |
 | MTLoBook3 |    | ConsLoBook3   |   |
 +-----------+    +---------------+   |
 +-----------+  +-| Book8 first   |   |
                | | ILoBook3 rest |---+
                | +---------------+ 
                |
                v
       +---------------+
       | Book8         |
       +---------------+
       | String title  |
       | String author |
       | int year      |
       +---------------+
*/

// to represent a list of books
interface ILoBook3 {

  // produce a list sorted by the titles from this list of books
  ILoBook3 sortByTitle();

  // insert the given book into this sorted list of books - by title
  ILoBook3 insertByTitle(Book8 b);

 // produce a list sorted by the year from this list of books
  ILoBook3 sortByYear();

  // insert the given book into this sorted list of books - by year
  ILoBook3 insertByYear(Book8 b);
}

// to represent an empty list of books
class MTLoBook3 implements ILoBook3 {

  MTLoBook3() {
  }

  // produce a list sorted by the titles from this list of books
  public ILoBook3 sortByTitle(){ return this; }

  // insert the given book into this sorted list of books - by title
  public ILoBook3 insertByTitle(Book8 b){ 
    return new ConsLoBook3(b, this);
  }

  // produce a list sorted by the year from this list of books
  public ILoBook3 sortByYear(){ return this; }

  // insert the given book into this sorted list of books - by year
  public ILoBook3 insertByYear(Book8 b){ 
    return new ConsLoBook3(b, this);
  }
}

// to represent a nonempty list of Books
class ConsLoBook3 implements ILoBook3 {
  Book8 first;
  ILoBook3 rest;

  ConsLoBook3(Book8 first, ILoBook3 rest) {
    this.first = first;
    this.rest = rest;
  }

  /* TEMPLATE:
    ... this.first ...                       -- Book8
    ... this.rest ...                        -- ILoBook3
    ... this.rest.sortByTitle() ...          -- ILoBook3 (sorted)
    ... this.rest.insertByTitle(Book) ...    -- ILoBook3 (sorted)
    ... this.rest.sortByYear() ...           -- ILoBook3 (sorted)
    ... this.rest.insertByYear(Book) ...     -- ILoBook3 (sorted)
  */

  // produce a list sorted by the titles from this list of books
  public ILoBook3 sortByTitle(){  
    return this.rest.sortByTitle().insertByTitle(this.first);
  }

  // insert the given book into this sorted list of books - by title
  public ILoBook3 insertByTitle(Book8 b){
    if (this.first.beforeByTitle(b))
      return new ConsLoBook3(this.first, this.rest.insertByTitle(b));
    else
      return new ConsLoBook3(b, this);
  }

  // produce a list sorted by the year from this list of books
  public ILoBook3 sortByYear(){  
    return this.rest.sortByYear().insertByYear(this.first);
  }

  // insert the given book into this sorted list of books - by year
  public ILoBook3 insertByYear(Book8 b){
    if (this.first.beforeByYear(b))
      return new ConsLoBook3(this.first, this.rest.insertByYear(b));
    else
      return new ConsLoBook3(b, this);
  }
}

// to represent a book
class Book8 {
  String title;
  String author;
  int year;

  Book8(String title, String author, int year) {
    this.title = title;
    this.author = author;
    this.year = year;
  }
/*
  // ** DRAFT TEMPLATE ** Edit as needed.
  ??? mmm() {
    ... this.title ...    -- String
    ... this.author ...   -- String
    ... this.year ...     -- int
  }
*/

  // is the title of this book before the title of the given book
  boolean beforeByTitle(Book8 that){
    return this.title.compareTo(that.title) < 0;
  }

  // is the year of this book before the year of the given book
  boolean beforeByYear(Book8 that){
    return this.year < that.year;
  }
}

// examples of data and tests
class ExamplesBook8{
  ExamplesBook8() {}

  Book8 b1 = new Book8("RC", "DD", 1796);
  Book8 b2 = new Book8("LPP", "StEx", 1942);
  Book8 b3 = new Book8("HtDP", "FFFK", 2001);

  ILoBook3 mtbooks = new MTLoBook3();
  ILoBook3 books1 = new ConsLoBook3(this.b1, 
                   new ConsLoBook3(this.b2, 
                   this.mtbooks));
  ILoBook3 books2 = new ConsLoBook3(this.b3, this.books1);

  ILoBook3 bSorted = new ConsLoBook3(this.b3, 
                    new ConsLoBook3(this.b1, 
                    this.mtbooks));

  ILoBook3 allSorted = new ConsLoBook3(this.b3, 
                      new ConsLoBook3(this.b2, 
                      new ConsLoBook3(this.b1, 
                      this.mtbooks)));

  // tests for the method beforeByTitle in the class Book
  boolean testBeforeByTitle(Tester t){
	  return
	  t.checkExpect(this.b2.beforeByTitle(this.b1), true) &&
	  t.checkExpect(this.b2.beforeByTitle(this.b3), false);
  }

  // tests for the method insertByTitle for ILoBook3 classes
  boolean testInsertByTitle(Tester t){
	  return
	  t.checkExpect(this.mtbooks.insertByTitle(this.b1),
         new ConsLoBook3(this.b1, this.mtbooks)) &&
	  t.checkExpect(this.bSorted.insertByTitle(this.b2),
         new ConsLoBook3(this.b3,
           new ConsLoBook3(this.b2,
             new ConsLoBook3(this.b1,
               this.mtbooks))));
  }

  // tests for the method sortByTitle for ILoBook3 classes
  boolean testSortByTitle(Tester t){
	  return
	  t.checkExpect(this.mtbooks.sortByTitle(), this.mtbooks) &&
	  t.checkExpect(this.books2.sortByTitle(), this.allSorted);
  }


  // tests for the method beforeByYear in the class Book
  boolean testBeforeByYear(Tester t){
	  return
	  t.checkExpect(this.b2.beforeByYear(this.b1), false) &&
	  t.checkExpect(this.b2.beforeByYear(this.b3), true);
  }

  // tests for the method insertByYear for ILoBook3 classes
  boolean testInsertByYear(Tester t){
	  return
	  t.checkExpect(this.mtbooks.insertByYear(this.b1),
          new ConsLoBook3(this.b1, this.mtbooks)) &&
      t.checkExpect(this.books1.insertByYear(this.b3),
          new ConsLoBook3(this.b1,
            new ConsLoBook3(this.b2,
              new ConsLoBook3(this.b3,
                this.mtbooks))));  
  }

// tests for the method sortByYear for ILoBook3 classes
  boolean testSortByYear(Tester t){
	  return
	  t.checkExpect(this.mtbooks.sortByYear(), this.mtbooks) &&
      t.checkExpect(this.books2.sortByYear(),
          new ConsLoBook3(this.b1, 
            new ConsLoBook3(this.b2, 
              new ConsLoBook3(this.b3, 
                this.mtbooks))));
  }

  // run the tests
  public static void main(String[] argv){
	  ExamplesBook8 eb8 = new ExamplesBook8();
	  Tester.runFullReport(eb8);
  }  
}
