// CS 2510 Spring 2010

import tester.*;
import draw.*;
import geometry.*;
import colors.*;

/*
  Sample mobiles:
  ---------------
  
  Simple mobile      |
                     |
                     2
                    blue
  
  Complex mobile         |
                         |
                         |
             ------------+-----
             |                |
       ------+------          |
       |           |          | 
       2           |          8    
      red          2        green
                  blue
*/


/*
               +---------+                  
               | IMobile |<---------------+
               +---------+                |
               +---------+                |
                   |                      |
                  / \                     |
                  ---                     |
                   |                      |
         ---------------------            |
         |                   |            |
  +--------------+   +---------------+    |
  | Simple       |   | Complex       |    |
  +--------------+   +---------------+    |
  | int length   |   | int length    |    |
  | int weight   |   | int leftside  |    |
  | IColor color |   | int rightside |    |
  +--------------+   | IMobile left  |----+ 
                     | IMobile right |----+  
                     +---------------+ 
*/

// to represent a mobile
interface IMobile {

  // count the weights in this mobile
  public int countWeights();

  // compute the total weight of this mobile
  public int totalWeight();
  
  // draw this mobile on the given canvas, hung at the given point
  public boolean drawMobile(Canvas c, Posn loc);
  
  // compute the height of this mobile
  //public int height();
  
  // is this mobile is balanced?
  //public boolean isBalanced();

}

// to represent an item hanging at the end of a mobile
class Simple implements IMobile {
  int length;
  int weight;
  IColor color;

  Simple(int length, int weight, IColor color) {
    this.length = length;
    this.weight = weight;
    this.color = color;
  }

/* TEMPLATE:
   FIELDS:
    ... this.length ...         -- int
    ... this.weight ...         -- int
    ... this.color ...          -- IColor

    METHODS:
    ... this.countWeights() ...     -- int
    ... this.totalWeight() ...      -- int
    ... draw(Canvas, Posn) ...      -- boolean
    ... this.height() ...           -- int
    ... this.isBalanced() ...       -- boolean
*/

  // count the weights in this simple mobile
  public int countWeights(){
    return 1;
  }
  // compute the total weight of this simple mobile
  // ... finish designing this method ...
  public int totalWeight(){
    return 0;
  }
  
  // draw this mobile on the given canvas, hung at the given point
  public boolean drawMobile(Canvas c, Posn loc){
	  return
	  c.drawLine(loc, new Posn(loc.x, loc.y + 5 * this.length), this.color) &&
	  c.drawDisk(new Posn(loc.x, loc.y + 5 * this.length - this.weight), 2 * this.weight, this.color);
  }
  
}

// to represent a part of support structure for a mobile
class Complex implements IMobile {
  int length;
  int leftSide;
  int rightSide;
  IMobile left;
  IMobile right;

  Complex(int length, int leftSide, int rightSide, 
          IMobile left, IMobile right) {
    this.length = length;
    this.leftSide = leftSide;
    this.rightSide = rightSide;
    this.left = left;
    this.right = right;
  }

/* TEMPLATE:
   FIELDS:
    ... this.length ...           -- int
    ... this.leftSide ...         -- int
    ... this.rightSide ...        -- int
    ... this.left ...             -- IMobile
    ... this.right ...            -- IMobile

    METHODS:
    ... this.countWeights() ...        -- int
    ... this.totalWeight() ...         -- int
    ... drawMobile(Canvas, Posn) ...   -- boolean
    ... this.height() ...              -- int
    ... this.isBalanced() ...          -- boolean
  
    METHODS FOR FIELDS:
    ... this.left.countWeights() ...           -- int
    ... this.left.totalWeight() ...            -- int
    ... this.left.drawMobile(Canvas, Posn) ... -- boolean
    ... this.left.height() ...                 -- int
    ... this.left.isBalanced() ...             -- boolean

    ... this.right.countWeights() ...           -- int
    ... this.right.totalWeight() ...            -- int
    ... this.right.drawMobile(Canvas, Posn) ... -- boolean
    ... this.right.height() ...                 -- int
    ... this.right.isBalanced() ...             -- boolean
*/

  // compute the total weight of this complex mobile
  public int countWeights(){
    return this.left.countWeights() + 
           this.right.countWeights();
  }
  
  // compute the total weight of this complex mobile
  // ... finish designing this method ...
  public int totalWeight(){
    return 0;
  }
  
  // draw this mobile on the given canvas, hung at the given point
  public boolean drawMobile(Canvas c, Posn loc){
	  // the left end of the strut
	  Posn leftEnd = new Posn(loc.x - 10 * this.leftSide, loc.y + 5 * this.length);
	  // the right end of the strut
	  Posn rightEnd = new Posn(loc.x + 10 * this.rightSide, loc.y + 5 * this.length);
	  // the center of the strut
	  Posn center = new Posn(loc.x, loc.y + 5 * this.length);
	  
	  return
	  // draw the line down
	  c.drawLine(loc, center, new Black()) &&
	  // draw the whole strut
	  c.drawLine(leftEnd, rightEnd, new Black()) &&
	  // draw the left mobile hanging from the left end of the strut
	  this.left.drawMobile(c, leftEnd) &&
	  // draw the right mobile hanging from the right end of the strut
	  this.right.drawMobile(c, rightEnd);
  }
 
}

/*
  Sample mobiles:
  ---------------
  
  simple1:
                     |
                     |
                     2
                    blue
  
  complex2 (with complex1 on the left and simple2 on the right)
                 complex1 (with simple3 on the left, simple1 on the right)
  
                         |
                         |
                         |
             ------------+------
             |                 |
       ------+------           |
       |           |           | 
       2           |           8    
      red          2          green
                  blue
*/
class ExamplesMobiles {
  ExamplesMobiles(){}

  IMobile simple1 = new Simple(20, 2, new Blue());
  IMobile simple2 = new Simple(30, 8, new Green());
  IMobile simple3 = new Simple(10, 2, new Red());

  IMobile complex1 = new Complex(10, 5, 5, this.simple3, this.simple1);
  IMobile complex2 = new Complex(30, 10, 5, this.complex1, this.simple2);
  
  // test the method totalWeight for the IMobile class hierarchy
  boolean testCountWeights(Tester t){
    return
    t.checkExpect(this.simple1.countWeights(), 1) &&
    t.checkExpect(this.complex1.countWeights(), 2) &&
    t.checkExpect(this.complex2.countWeights(), 3);
  }
  
  boolean testDrawing(Tester t){
	  Canvas c = new Canvas(400, 600);
	  return 
	  t.checkExpect(c.show() && this.complex2.drawMobile(c, new Posn(200, 20)), true);
  }
	
	// run the program/tests directly
	public static void main(String[] argv){
	  ExamplesMobiles e = new ExamplesMobiles();
	  Tester.runReport(e, true, false);
	}
}
