import colors.*;
import draw.*;
import geometry.*;

import java.util.Random;
import tester.*;

// to represent a shark in an ocean world
class Shark {
  Posn loc;
  int life;
  
  Shark(Posn loc, int life) {
    this.loc = loc;
    this.life = life;
  }
  
  // produce a new shark from this shark moved in response to the key press
  Shark onKeyEvent(String ke){
    if (ke.equals("up")){
      return new Shark(new Posn(this.loc.x, this.loc.y - 3), this.life);
    }
    else {if (ke.equals("down")){
      return new Shark(new Posn(this.loc.x, this.loc.y + 3), this.life);
    }
    else{
      return this;
    }}
  }
  
  // draw this shark on the given Canvas
  boolean draw(Canvas c){
    return c.drawDisk(this.loc, this.life / 10 + 4, new Black()) &&
           c.drawRect(new Posn(this.loc.x + this.life / 10 + 1, this.loc.y - 3), 10, 6, new Black()) &&
           c.drawLine(new Posn(this.loc.x + this.life / 10 + 1, this.loc.y - 1),
                      new Posn(this.loc.x + this.life / 10 + 10, this.loc.y - 1), 
                      new Yellow()) &&
           c.drawLine(new Posn(this.loc.x + this.life / 10 + 1, this.loc.y),
                      new Posn(this.loc.x + this.life / 10 + 10, this.loc.y), 
                      new Yellow()) &&
           this.drawEye(c);
  }
  
  // draw shark's eye - if he is big enough
  boolean drawEye(Canvas c){
    if (this.life > 20){
      return c.drawDisk(new Posn(this.loc.x + this.life / 10 - 3, this.loc.y - 2), 2, new White());
    }
    else{
      return true;
    }
  }
  
  // produce a shark hungrier by a minute than this one
  Shark onTick(){
    if (this.life > 0){
      return new Shark(this.loc, this.life - 1);
    }
    else {
      return this;
    }
  }
    
  // produce a shark after this one ate one fish
  // for now every fish tastes the same
  // later, fish nutrition value may be determined by the size or the color 
  Shark getFatter(){
    return new Shark(this.loc, this.life + 30);
  }
  
  // is this shark dead?
  boolean isDead(){
    return this.life <= 0;
  }  
}

// to represent a fish in an ocean world
class Fish {
  Posn loc;
  boolean hasEscaped;
  static Random random = new Random();

  Fish(Posn loc, boolean hasEscaped) {
    this.loc = loc;
    this.hasEscaped = hasEscaped;
  }
  
  // produce from this fish the one that swam for a minute more
  Fish onTick(){
    if (this.hasEscaped){
      return this.makeFish(600, 400);
    }
    else {if (this.loc.x < 0){
      return new Fish(this.loc, true);
    }
    else {
      return new Fish(new Posn(this.loc.x - 3, 
                               this.loc.y + random.nextInt(7) - 3),
                               false);
    }}
  }
  
  // has this fish escaped?
  boolean escaped(){
    return (this.loc.x < 0) || this.hasEscaped;
  }
  
  // is this fish a food for the given shark?
  boolean isFood(Shark shark){
    return
      Math.sqrt((this.loc.x - shark.loc.x) * (this.loc.x - shark.loc.x) +
                (this.loc.y - shark.loc.y) * (this.loc.y - shark.loc.y)) < 25;     
  }

  // draw this fish on the given Canvas
  boolean draw(Canvas c){
    return c.drawDisk(this.loc, 10, new Red()) &&
           c.drawDisk(new Posn(this.loc.x - 3, this.loc.y), 2, new White()) &&
           c.drawRect(new Posn(this.loc.x + 8, this.loc.y - 4), 4, 9, new Red());
  }
  
  // produce a new fish at the right quarter of the canvas at random height
  Fish makeFish(int WIDTH, int HEIGHT){
    return new Fish(new Posn(WIDTH - WIDTH / 4 + this.randomLimited(WIDTH / 4), 
                    this.randomLimited(HEIGHT)), false);
  }
    
  // produce a random initial coordinate of the fish
  int randomLimited(int LIMIT){
    return new Random().nextInt() % LIMIT;
  }
  
  // replace this fish with a new one
  Fish feedShark(Shark shark){
    return this.makeFish(600, 400); 
  }
}

// to represent a list of Fish
interface ILoFish {
  // produce from this fish the one that swam for a minute more
  ILoFish  onTick();
  
  // is there a fish in this school that can be eaten by the given shark?
  boolean isFood(Shark shark);

  // produce a school of fish with one eaten and replaced by a new random one
  ILoFish feedShark(Shark shark);

  // draw this fish on the given Canvas
  boolean draw(Canvas c);
}

// to represent an empty list of Fish
class MtLoFish implements ILoFish {

  MtLoFish() {
  }
  
  // produce from this fish the one that swam for a minute more
  public ILoFish onTick(){
    return this;
  }
  
  // is there a fish in this school that can be eaten by the given shark?
  public boolean isFood(Shark shark){
    return false;
  }

  // produce a school of fish with one eaten and replaced by a new random one
  public ILoFish feedShark(Shark shark){
    return this;
  }

  // draw this fish on the given Canvas
  public boolean draw(Canvas c){
    return true;
  }
}

// to represent a nonempty list of Fish
class ConsLoFish implements ILoFish {
  Fish first;
  ILoFish rest;

  ConsLoFish(Fish first, ILoFish rest) {
    this.first = first;
    this.rest = rest;
  }

  // produce from this fish the one that swam for a minute more
  public ILoFish  onTick(){
    return new ConsLoFish(this.first.onTick(),
                          this.rest.onTick()); 
  }
  
  // is there a fish in this school that can be eaten by the given shark?
  public boolean isFood(Shark shark){
    return this.first.isFood(shark) || this.rest.isFood(shark);
  }

  // produce a school of fish with one eaten and replaced by a new random one
  public ILoFish feedShark(Shark shark){
    if (this.first.isFood(shark)){
      return new ConsLoFish(this.first.makeFish(200, 200), this.rest);
    }
    else {
      return new ConsLoFish(this.first, this.rest.feedShark(shark));
    }
  }

  // draw this fish on the given Canvas
  public boolean draw(Canvas c){
    return this.first.draw(c) &&
           this.rest.draw(c);
  }
}

// to represent an ocean world
class OceanWorld extends World{
  Shark shark;
  ILoFish fish;
  int WIDTH =  600;
  int HEIGHT = 400;

  OceanWorld(Shark shark, ILoFish fish) {
    this.shark = shark;
    this.fish = fish;
  }
  
  // start the world and the timer
  boolean go() { return this.bigBang(600, 400, 0.05); }

  // produce a new OceanWorld after one minute elapsed: 
  // move the fish, starve the shark, check if the fish is eaten or has escaped
  public World onTick(){
    // if the shark found fish, fed the shark, replace the fish with a new one                        
    if (this.fish.isFood(this.shark)){
      return new OceanWorld(this.shark.getFatter(),
                            this.fish.feedShark(shark));
    }
                            
    // if the shark starved to death, end the world
    else {if(this.shark.isDead()) {
      return this.endOfWorld("The shark starved to death");
    }
    
    // no special events, just move the fish and starve the shark 
    else {
      return new OceanWorld(this.shark.onTick(), this.fish.onTick());
    }}
  }
  
  // produce a new OceanWorld in response to the given key press
  public World onKeyEvent(String ke){
    return new OceanWorld(this.shark.onKeyEvent(ke), this.fish);
  }
  
  // draw this world
  public boolean draw(){
    return this.theCanvas.drawRect(new Posn(0, 0), 
                            this.WIDTH, this.HEIGHT, new Blue()) &&
           this.fish.draw(this.theCanvas) &&
           this.shark.draw(this.theCanvas);
  }
}

// Examples of the classes Fish, Shark, OceanWorld and tests for all methods
class ExamplesOceanWorld{
	ExamplesOceanWorld(){}
  
  Shark mac = new Shark(new Posn(10, 100), 200);
  Shark mac2 = new Shark(new Posn(10, 100), 100);
  Shark mac3 = new Shark(new Posn(10, 100), -1);
  
  Fish fishy = new Fish(new Posn(190, 100), false);
  Fish yummy = new Fish(new Posn(15, 98), false);
  Fish gone = new Fish(new Posn(15, 98), true);
  Fish gone2 = new Fish(new Posn(-2, 98), false);
  
  
  Fish fishy1 = this.fishy.makeFish(600, 400);
  Fish fishy2 = this.fishy.makeFish(600, 400);
  Fish fishy3 = this.fishy.makeFish(600, 400);
  Fish fishy4 = this.fishy.makeFish(600, 400);
  Fish fishy5 = this.fishy.makeFish(600, 400);
  Fish fishy6 = this.fishy.makeFish(600, 400);
  Fish fishy7 = this.fishy.makeFish(600, 400);
  Fish fishy8 = this.fishy.makeFish(600, 400);
  Fish fishy9 = this.fishy.makeFish(600, 400);
  Fish fishy10 = this.fishy.makeFish(600, 400);
  Fish fishy11 = this.fishy.makeFish(600, 400);
  Fish fishy12 = this.fishy.makeFish(600, 400);
  Fish fishy13 = this.fishy.makeFish(600, 400);
  Fish fishy14 = this.fishy.makeFish(600, 400);
  Fish fishy15 = this.fishy.makeFish(600, 400);
  
  ILoFish nofish = new MtLoFish();

  ILoFish fishies = new ConsLoFish(this.fishy1,
                    new ConsLoFish(this.fishy2,
                    new ConsLoFish(this.fishy3,
                    new ConsLoFish(this.fishy4,
                    new ConsLoFish(this.fishy5,new MtLoFish())))));
                    		
  ILoFish fishLot = new ConsLoFish(this.fishy1,
                    new ConsLoFish(this.fishy2,
                    new ConsLoFish(this.fishy3,
                    new ConsLoFish(this.fishy4,
                    new ConsLoFish(this.fishy5,
                    new ConsLoFish(this.fishy6,
                    new ConsLoFish(this.fishy7,
                    new ConsLoFish(this.fishy8,
                    new ConsLoFish(this.fishy9,
                    new ConsLoFish(this.fishy10,
                    new ConsLoFish(this.fishy11,
                    new ConsLoFish(this.fishy12,
                    new ConsLoFish(this.fishy13,
                    new ConsLoFish(this.fishy14,
                    new ConsLoFish(this.fishy15,new MtLoFish())))))))))))))));
                    
  ILoFish manyfish = new ConsLoFish(this.fishy,
                     new ConsLoFish(this.yummy, new MtLoFish()));
                    
    
  OceanWorld sfw = new OceanWorld(this.mac, this.fishies); 
  OceanWorld sfwBig = new OceanWorld(this.mac, this.fishLot); 
  
  
  Canvas c = new Canvas(200, 200);
  
  boolean go(){
    return sfwBig.bigBang(600, 400, 0.1);
  }
  
  // test the method onKeyEvent in the class Shark
  boolean testSharkOnKeyEvent(Tester t){
    return
      t.checkExpect(this.mac.onKeyEvent("left"), this.mac) && 
      t.checkExpect(this.mac.onKeyEvent("up"),
        new Shark(new Posn(10, 97), 200)) && 
      t.checkExpect(this.mac.onKeyEvent("down"),
        new Shark(new Posn(10, 103), 200));
  }
  
  // a visual test for drawing a Shark
  boolean testDrawShark(Tester t){
    return this.c.show() && this.mac.draw(this.c);
  }
  
  // test the method onTick in the class Shark
  boolean testSharkOnTick(Tester t){
    return
    t.checkExpect(this.mac.onTick(),
        new Shark(new Posn(10, 100), 199)) && 
    t.checkExpect((new Shark(new Posn(10, 100), 0)).onTick(),
        new Shark(new Posn(10, 100), 0)); 
  }
  
  // test the method getFatter in the class Shark
  boolean testGetFatter(Tester t){
    return
    t.checkExpect(this.mac.getFatter(),
        new Shark(new Posn(10, 100), 230)) &&
    t.checkExpect(this.mac2.getFatter(),
        new Shark(new Posn(10, 100), 130));
  }

  // test the method isDead in the class Shark
  boolean testisDead(Tester t){
    return
    t.checkExpect(this.mac.isDead(), false) &&
    t.checkExpect(this.mac3.isDead(), true);
  }


  // test the method onTick in the class Fish
  // results are correct if the random move is replaced with exactly 5
  boolean testFishOnTick(Tester t){
    return 
    t.checkExpect(this.fishy.onTick(), 
        new Fish(new Posn(187, 105), false)) &&
    t.checkExpect((new Fish(new Posn(-2, 100), true)).onTick(),
        new Fish(new Posn(-2, 100), true)) && 
    t.checkExpect((new Fish(new Posn(-2, 100), false)).onTick(),
        new Fish(new Posn(-2, 100), true)); 
  }
  
  // test the method escaped in the class Fish
  boolean testFishEscaped(Tester t){
    return 
    t.checkExpect(this.fishy.escaped(), false) && 
    t.checkExpect((new Fish(new Posn(-2, 100), true)).escaped(), true) &&
    t.checkExpect((new Fish(new Posn(-2, 100), false)).escaped(), true); 
  }

  // test the method isFood in the class Fish
  boolean testFishIsFood(Tester t){
    return 
    t.checkExpect(this.fishy.isFood(this.mac), false) &&
    t.checkExpect(this.yummy.isFood(new Shark(new Posn(10, 100), 20)), true);
  } 

  // a visual test for drawing a Fish
  boolean drawFish(Tester t){
    return this.c.show() && this.fishy.draw(this.c);
  }
  
  // test for the method randomLimited in the class Fish
  boolean testRandomLimited(Tester t){
    return 
    t.checkExpect(this.fishy.randomLimited(5) < 6 && 
                  this.fishy.randomLimited(5) > 0, true);
  }

  // test the method makeFish in the class Fish
  boolean testMakeFish(Tester t){
    return 
    t.checkExpect(this.fishy.makeFish(600, 400).hasEscaped, false) &&
    t.checkExpect(this.fishy.makeFish(600, 400).loc.y < 400, true) &&
    t.checkExpect(this.fishy.makeFish(600, 400).loc.y >   0, true) &&
    t.checkExpect(this.fishy.makeFish(600, 400).loc.x < 600, true) &&
    t.checkExpect(this.fishy.makeFish(600, 400).loc.x > 550, true);
  }
  


  ILoFish manyfishNext = new ConsLoFish(this.fishy.onTick(),
                         new ConsLoFish(this.yummy.onTick(), new MtLoFish()));
  ILoFish manyfishEaten = new ConsLoFish(this.fishy,
                          new ConsLoFish(this.yummy.feedShark(this.mac), new MtLoFish()));

  // test the method onTick in the classes that represent a school of fish
  boolean testILoFishOnTick(Tester t){
    return 
    t.checkExpect(this.nofish.onTick(), this.nofish) &&
    t.checkExpect(this.manyfish.onTick(), this.manyfishNext);
  }

  // test the method isFood in the classes that represent a school of fish
  boolean testILoFishIsFood(Tester t){
    return 
    t.checkExpect(this.nofish.isFood(this.mac), false) &&
    t.checkExpect(this.manyfish.isFood(this.mac), true) &&
    t.checkExpect(this.fishies.isFood(this.mac), false);
  }

  // test the method feedShark in the classes that represent a school of fish
  boolean testILoFishFeedShark(Tester t){
    return 
    t.checkExpect(this.nofish.feedShark(this.mac), this.nofish) &&
    t.checkExpect(this.manyfish.feedShark(this.mac), this.manyfishEaten) &&
    t.checkExpect(this.fishies.feedShark(this.mac), this.fishies);
  }

  // a visual test for drawing a school of fish
  boolean drawFishies(Tester t){
    return this.c.show() && this.manyfish.draw(this.c);
  }

  // test the method onKeyEvent in the class OceanWorld
  boolean testOceanOnKeyEvent(Tester t){
    return
    t.checkExpect(this.sfw.onKeyEvent("left"),
        new OceanWorld(this.mac, this.fishies)) && 
    t.checkExpect(this.sfw.onKeyEvent("up"), 
        new OceanWorld(new Shark(new Posn(10, 97), 200), this.fishies)) && 
    t.checkExpect(this.sfw.onKeyEvent("down"),
        new OceanWorld(new Shark(new Posn(10, 103), 200), this.fishies));
  }
  
  OceanWorld sfwEnd = new OceanWorld(this.mac3, this.fishies);
  OceanWorld sfwReg = new OceanWorld(this.mac, this.fishies);
  OceanWorld sfwYum = new OceanWorld(this.mac, this.manyfish);

  // test the method onTick in the class OceanWorld
  boolean testOceanOnTick(Tester t){
    return
      // world goes on
    t.checkExpect(this.sfwReg.onTick(), 
        new OceanWorld(this.mac.onTick(), this.fishies.onTick())) &&
        
      // fish is eaten, shark is fatter, new fish appears
    t.checkExpect(this.sfwYum.onTick() ,
        new OceanWorld(this.mac.getFatter(), 
                              this.manyfishEaten));
      // world ends - the shark starves to death --- fails due to randomness
      // cannot invoke method in the teachpack - no test is possible
      //(check this.sfwEnd.onTick() 
      //  expect this.sfw.endWorld("The shark has starved to death"));
  }

  // a visual test for drawing the OceanWorld
  boolean drawOcean(Tester t){
    return this.sfw.draw();
  }

  
  /** main: an alternative way of starting the world and running the tests */
  public static void main(String[] argv){
    // test the methods for the OceanWorld
    ExamplesOceanWorld ow = new ExamplesOceanWorld();
    Tester.run(ow);
    
    // run the OceanWorld
    ow.go();
  }
}